/***************************************************************** includes ***/
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "uart.h"

/****************************************************************** defines ***/
#ifndef F_CPU
#define F_CPU 4000000UL
#endif

#define UART_BAUD_RATE  9600

/* Communication flow control */
#define XON             17
#define XOFF            19

/* Start symbol of rows in hex file*/
#define START_SIGN      ':'

/* States of the bootloader-program*/
#define BOOT_STATE_EXIT   0
#define BOOT_STATE_PARSER 1

/* State of the hex-file-parser */
#define PARSER_STATE_START    0
#define PARSER_STATE_SIZE     1
#define PARSER_STATE_ADDRESS  2
#define PARSER_STATE_TYPE     3
#define PARSER_STATE_DATA     4
#define PARSER_STATE_CHECKSUM 5
#define PARSER_STATE_ERROR    6

/*************************************************************** prototypes ***/
void fnProgramPage(uint32_t page, uint8_t *buffer);
static uint16_t fnHex2Num(const uint8_t* cASCII, uint8_t nNum);

/************************************************************ main-function ***/
int main(void) {

	/* Character form UART + status code */
	uint16_t c = 0;
	/* Intel-hex target adress */
	uint16_t hex_addr = 0;
	/* Flash page to write on */
	uint16_t flash_page = 0;
	/* Intel-hex checksum */
	uint16_t hex_check = 0;
	/* Position to write on */
	uint16_t flash_cnt = 0;
	/* Temporary variable */
	uint8_t temp;
	/* Flag for programming mode control */
	uint8_t boot_state = BOOT_STATE_EXIT;
	/* Flow control flag */
	uint8_t parser_state = PARSER_STATE_START;
	/* Flag for of new flash page */
	uint8_t flash_page_flag = 1;
	/* Buffer for hex-files */
	uint8_t flash_data[SPM_PAGESIZE];
	/* Position to write in the hex buffer */
	uint8_t hex_cnt = 0;
	/* Buffer for conversion from ascii to binary */
	uint8_t hex_buffer[5];
	/* Intel-hex data length */
	uint8_t hex_size = 0;
	/* Counter for received hex data per line */
	uint8_t hex_data_cnt = 0;
	/* Intel-hex recordtype */
	uint8_t hex_type = 0;
	/* received hex checksum */
	uint8_t hex_checksum = 0;

	void(*start)(void)=0x0000;

	/* Filling buffer with defined values */
	memset(hex_buffer, 0x00, sizeof(hex_buffer));
	memset(flash_data, 0xFF, sizeof(flash_data));

	/* Edit interrupt vectors */
	temp = MCUCR;
	MCUCR = temp | (1 << IVCE);
	MCUCR = temp | (1 << IVSEL);

	/* Initialize UART */
	uart_init(UART_BAUD_SELECT(UART_BAUD_RATE, F_CPU));
	sei();

	uart_puts_P("\nBOOTLOADER FOR ATMEL AVR ATMEGA644\n\r");
	uart_puts_P("Version: 1.0\n\n\r");
	_delay_ms(2000);

	do {
		c = uart_getc();
		if (!(c & UART_NO_DATA)) {
			if (c & UART_FRAME_ERROR)
				uart_puts_P("ERROR: Framing error detected!\n\r");
			if (c & UART_OVERRUN_ERROR)
				uart_puts_P("ERROR: Overrun error detected!\n\r");
			if (c & UART_BUFFER_OVERFLOW)
				uart_puts_P("ERROR: UART buffer overflow!\n\r");
			/* Programming mode parser */
			if (boot_state == BOOT_STATE_PARSER) {
				switch (parser_state) {
				/* Waiting for line start character */
				case PARSER_STATE_START:
					if ((uint8_t) c == START_SIGN) {
						uart_putc(XOFF);
						parser_state = PARSER_STATE_SIZE;
						hex_cnt = 0;
						hex_check = 0;
						uart_putc(XON);
					}
					break;
					/* Parse datasize */
				case PARSER_STATE_SIZE:
					hex_buffer[hex_cnt++] = (uint8_t) c;
					if (hex_cnt == 2) {
						uart_putc(XOFF);
						parser_state = PARSER_STATE_ADDRESS;
						hex_cnt = 0;
						hex_size = (uint8_t) fnHex2Num(hex_buffer, 2);
						hex_check += hex_size;
						uart_putc(XON);
					}
					break;
					/* Parse target address */
				case PARSER_STATE_ADDRESS:
					hex_buffer[hex_cnt++] = (uint8_t) c;
					if (hex_cnt == 4) {
						uart_putc(XOFF);
						parser_state = PARSER_STATE_TYPE;
						hex_cnt = 0;
						hex_addr = fnHex2Num(hex_buffer, 4);
						hex_check += (uint8_t) hex_addr;
						hex_check += (uint8_t) (hex_addr >> 8);
						if (flash_page_flag) {
							flash_page = hex_addr - hex_addr % SPM_PAGESIZE;
							flash_page_flag = 0;
						}
						uart_putc(XON);
					}
					break;
					/* Parse linetype */
				case PARSER_STATE_TYPE:
					hex_buffer[hex_cnt++] = (uint8_t) c;
					if (hex_cnt == 2) {
						uart_putc(XOFF);
						hex_cnt = 0;
						hex_data_cnt = 0;
						hex_type = (uint8_t) fnHex2Num(hex_buffer, 2);
						hex_check += hex_type;
						switch (hex_type) {
						case 0:
							parser_state = PARSER_STATE_DATA;
							break;
						case 1:
							parser_state = PARSER_STATE_CHECKSUM;
							break;
						default:
							parser_state = PARSER_STATE_DATA;
							break;
						}
						uart_putc(XON);
					}
					break;
					/* Parse flash-data */
				case PARSER_STATE_DATA:
					hex_buffer[hex_cnt++] = (uint8_t) c;
					if (hex_cnt == 2) {
						uart_putc(XOFF);
						uart_putc('.');
						hex_cnt = 0;
						flash_data[flash_cnt] = (uint8_t) fnHex2Num(hex_buffer,
								2);
						hex_check += flash_data[flash_cnt];
						flash_cnt++;
						hex_data_cnt++;
						if (hex_data_cnt == hex_size) {
							parser_state = PARSER_STATE_CHECKSUM;
							hex_data_cnt = 0;
							hex_cnt = 0;
						}
						/* Buffer full, write page */
						if (flash_cnt == SPM_PAGESIZE) {
							uart_puts("P\n\r");
							_delay_ms(100);
							fnProgramPage((uint16_t) flash_page, flash_data);
							memset(flash_data, 0xFF, sizeof(flash_data));
							flash_cnt = 0;
							flash_page_flag = 1;
						}
						uart_putc(XON);
					}
					break;
					/* Parse checksum */
				case PARSER_STATE_CHECKSUM:
					hex_buffer[hex_cnt++] = (uint8_t) c;
					if (hex_cnt == 2) {
						uart_putc(XOFF);
						hex_checksum = (uint8_t) fnHex2Num(hex_buffer, 2);
						hex_check += hex_checksum;
						hex_check &= 0x00FF;
						/* file end, write rest of data */
						if (hex_type == 1) {
							uart_puts("P\n\r");
							_delay_ms(100);
							fnProgramPage((uint16_t) flash_page, flash_data);
							boot_state = BOOT_STATE_EXIT;
						}
						/* test if checksum is 0 */
						if (hex_check == 0)
							parser_state = PARSER_STATE_START;
						else
							parser_state = PARSER_STATE_ERROR;
						uart_putc(XON);
					}
					break;
					/* Parse error */
				case PARSER_STATE_ERROR:
					uart_putc('#');
					break;
				default:
					break;
				}
			}
			/* Programming mode UART communication */
			else if (boot_state != BOOT_STATE_PARSER) {
				switch ((uint8_t) c) {
				case 'p':
					boot_state = BOOT_STATE_PARSER;
					uart_puts_P("Programming flash.\n\r");
					uart_puts_P("Insert hex-file here:\n\r");
					break;
				case 'q':
					boot_state = BOOT_STATE_EXIT;
					uart_puts_P("Leaving bootloader.\n\r");
					break;
				default:
					uart_puts_P("USAGE: \n\r");
					uart_puts_P("      q ... quit\n\r");
					uart_puts_P("      p ... flash hex-file\n\r");
					break;
				}
			}
		}
	} while (boot_state != BOOT_STATE_EXIT);

	uart_puts("Reset AVR.\n\r");
	_delay_ms(1000);

	/* Restore interrupt vector */
	temp = MCUCR;
	MCUCR = temp | (1<<IVCE);
	MCUCR = temp & ~(1<<IVSEL);

	/* Reset */
	start();

	return 0;
}

/**************************************************************** functions ***/

void fnProgramPage(uint32_t unPage, uint8_t *acBuffer) {

	uint16_t i;
	uint8_t sreg;

	sreg = SREG;
	cli();

	eeprom_busy_wait();

	/* Erase Page and wait until finish */
	boot_page_erase(unPage);
	boot_spm_busy_wait();

	for (i = 0; i < SPM_PAGESIZE; i += 2) {

		/* Set up little endian word */
		uint16_t w = *acBuffer++;
		w += (*acBuffer) << 8;

		boot_page_fill(unPage+i, w);
	}

	/* Store buffer to flash page and wait until finish */
	boot_page_write(unPage);
	boot_spm_busy_wait();

	/* Reenable RWW-section. Needet to jump back to application */boot_rww_enable();

	/* Reenable interrupts */
	SREG = sreg;
}

static uint16_t fnHex2Num(const uint8_t* aASCII, uint8_t nNum) {
	uint8_t i;
	uint16_t nVal = 0;

	for (i = 0; i < nNum; i++) {
		uint8_t c = aASCII[i];

		/* Generate bin from hex */
		if (c >= '0' && c <= '9')
			c -= '0';
		if (c >= 'A' && c <= 'F')
			c -= 'A' - 10;
		if (c >= 'a' && c <= 'f')
			c -= 'a' - 10;
		nVal = 16 * c;
	}
	return nVal;
}
